import Negotiator from 'negotiator';

export interface MediaType<
	Name extends string = string,
	T extends object | null = object | null,
	SerializeOpts extends {} = {},
	DeserializeOpts extends {} = {}
> {
	name: Name;
	serialize: (object: T, args?: SerializeOpts) => string;
	deserialize: (s: string, args?: DeserializeOpts) => T;
}

export const FALLBACK_MEDIA_TYPE = {
	serialize: (obj: unknown) => JSON.stringify(obj),
	deserialize: (str: string) => JSON.parse(str),
	name: 'application/json' as const,
} satisfies MediaType;

export const PATCH_CONTENT_TYPES = [
	'application/merge-patch+json',
	'application/json-patch+json',
] as const;

export type PatchContentType = typeof PATCH_CONTENT_TYPES[number];

export const getAcceptPostString = (mediaTypes: Map<string, MediaType>) => Array.from(mediaTypes.keys())
	.filter((t) => !PATCH_CONTENT_TYPES.includes(t as PatchContentType))
	.join(',');

export const parseMediaType = (mediaType: string) => {
	const [type, subtypeEntirety] = mediaType.split('/').map((s) => s.trim());
	return {
		type,
		// application/foo+json
		subtype: subtypeEntirety.split('+'),
	};
};

export const isTextMediaType = (mediaType: string) => {
	const { type, subtype } = parseMediaType(mediaType);
	if (type === 'text') {
		return true;
	}

	if (type === 'application' && subtype.length > 0) {
		const lastSubtype = subtype.at(-1) as string;
		return [
			'json',
			'xml'
		].includes(lastSubtype);
	}

	return false;
};

export const getMediaType = (availableMediaTypes: string[], mediaTypeString?: string): MediaType['name'] | undefined => {
	if (typeof mediaTypeString === 'undefined') {
		return FALLBACK_MEDIA_TYPE['name'];
	}

	const negotiator = new Negotiator({
		headers: {
			'accept': mediaTypeString,
		},
	});
	return negotiator.mediaType(availableMediaTypes);
};

export const parseAcceptString = (acceptString?: string) => {
	if (typeof acceptString !== 'string') {
		return undefined;
	}
	const [mediaType, ...acceptParams] = acceptString.split(';');
	const params = Object.fromEntries(
		acceptParams.map((s) => {
			const [key, ...values] = s.split('=');
			return [key.trim(), ...values.map((v) => v.trim()).join('=')];
		})
	);

	return {
		mediaType,
		params,
	};
};
