import http from 'http';
import {
	ErrorResponse,
	parseToEndpointQueue,
	ServiceParams,
	statusCodes,
	getContentNegotiationParams,
	FALLBACK_LANGUAGE,
	FALLBACK_CHARSET,
	FALLBACK_MEDIA_TYPE,
	parseAcceptString,
} from '@modal-sh/yasumi';
import {
	Backend as BaseBackend,
	Server,
	ServerRequestContext,
	ServerResponseContext,
	ServerParams,
} from '@modal-sh/yasumi/backend';

declare module '@modal-sh/yasumi/backend' {
	interface ServerRequestContext extends http.IncomingMessage {}

	interface ServerResponseContext extends http.ServerResponse {}
}

const getBody = (
	req: http.IncomingMessage,
) => new Promise<Buffer>((resolve, reject) => {
	let body = Buffer.from('');
	req.on('data', (chunk) => {
		body = Buffer.concat([body, chunk]);
	});
	req.on('end', async () => {
		resolve(body);
	});

	req.on('error', (err) => {
		reject(err);
	});
});

class ServerInstance<Backend extends BaseBackend> implements Server<Backend> {
	readonly backend: Backend;
	private serverInternal?: http.Server;

	constructor(params: ServerParams<Backend>) {
		this.backend = params.backend;
	}

	private readonly requestListener = async (req: ServerRequestContext, res: ServerResponseContext) => {
		const { method: methodRaw, url, headers: requestHeaders } = req;
		const {
			language,
			mediaType,
			charset,
		} = getContentNegotiationParams(this.backend.app, requestHeaders);
		const fallbackLanguage = language ?? FALLBACK_LANGUAGE;
		// TODO use these for errors
		const fallbackMediaType = mediaType ?? FALLBACK_MEDIA_TYPE;
		const fallbackCharset = charset ?? FALLBACK_CHARSET;
		const errorHeaders = {
			'content-language': fallbackLanguage.name,
		} as Record<string, string>;

		if (typeof methodRaw === 'undefined') {
			res.writeHead(statusCodes.HTTP_STATUS_BAD_REQUEST, errorHeaders);
			res.end();
			return;
		}
		if (typeof url === 'undefined') {
			res.writeHead(statusCodes.HTTP_STATUS_BAD_REQUEST, errorHeaders);
			res.end();
			return;
		}

		const method = methodRaw.toUpperCase();
		console.log(language?.name, mediaType?.name, charset?.name);
		console.log(method, url);

		const endpoints = parseToEndpointQueue(url, this.backend.app.endpoints);
		const [endpoint, endpointParams] = endpoints.at(-1) ?? [];

		if (typeof endpoint === 'undefined') {
			res.writeHead(statusCodes.HTTP_STATUS_NOT_FOUND, errorHeaders);
			res.end();
			return;
		}

		const appOperations = Array.from(this.backend.app.operations.values())
		const foundAppOperation = appOperations.find((op) => {
			const doesMethodMatch = op.method === method;
			if (typeof op.headers !== 'undefined') {
				const doesHeadersMatch = Array.from(op.headers.entries()).reduce(
					(currentHeadersMatch, [headerKey, opHeaderValue]) => (
						// TODO honor content-type matching
						currentHeadersMatch && requestHeaders[headerKey.toLowerCase()] === opHeaderValue
					),
					true,
				);

				return (
					doesMethodMatch
					&& doesHeadersMatch
				);
			}

			return doesMethodMatch;
		});

		if (typeof foundAppOperation === 'undefined') {
			res.writeHead(statusCodes.HTTP_STATUS_METHOD_NOT_ALLOWED, {
				...errorHeaders,
				'Allow': appOperations.map((op) => op.method).join(',')
			});
			res.end();
			return;
		}

		if (!endpoint.operations.has(foundAppOperation.name)) {
			const endpointOperations = Array.from(endpoint?.operations ?? []);
			res.writeHead(statusCodes.HTTP_STATUS_METHOD_NOT_ALLOWED, {
				...errorHeaders,
				'Allow': endpointOperations
					.map((a) => appOperations.find((aa) => aa.name === a)?.method)
					.join(',')
			});
			res.end();
			return;
		}

		const implementation = this.backend.implementations.get(foundAppOperation.name);
		if (typeof implementation === 'undefined') {
			res.writeHead(statusCodes.HTTP_STATUS_NOT_IMPLEMENTED, errorHeaders);
			res.end();
			return;
		}

		const [, search] = url.split('?');

		// TODO add flag on implementation context if CQRS should be enabled

		try {
			// TODO deserialize body, else throw 415
			const requestBodySpecs = parseAcceptString(requestHeaders['content-type']);
			let body: unknown = undefined;
			// TODO check body
			if (typeof requestBodySpecs !== 'undefined') {
				const {
					mediaType: requestBodyMediaTypeString,
					params: requestBodyParams,
				} = requestBodySpecs;
				const requestBodyCharsetString = requestBodyParams?.charset;
				const requestBodyMediaType = this.backend.app.mediaTypes.get(requestBodyMediaTypeString);
				if (typeof requestBodyMediaType === 'undefined') {
					res.writeHead(statusCodes.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE, errorHeaders);
					res.end();
					return;
				}
				let decoder = (buf: Buffer, _params?: {}) => buf.toString('binary');
				if (typeof requestBodyCharsetString !== 'undefined') {
					const requestBodyCharset = this.backend.app.charsets.get(requestBodyCharsetString);
					if (typeof requestBodyCharset !== 'undefined') {
						decoder = requestBodyCharset.decode;
					}
				}

				const bodyBuffer = await getBody(req);
				// TODO set params to decoder/deserializer
				const bodyDecoded = decoder(bodyBuffer, requestBodyParams);
				body = requestBodyMediaType.deserialize(bodyDecoded, requestBodyParams);
			}
			const responseSpec = await implementation({
				endpoint,
				body,
				params: endpointParams ?? {},
				query: typeof search !== 'undefined' ? new URLSearchParams(search) : undefined,
				dataSource: this.backend.dataSource,
			});

			if (typeof responseSpec === 'undefined') {
				res.writeHead(statusCodes.HTTP_STATUS_UNPROCESSABLE_ENTITY, errorHeaders);
				res.end();
				return;
			}

			const responseHeaders = {} as Record<string, string>;
			// TODO serialize using content-negotiation params
			const bodyToSerialize = responseSpec.body;
			if (bodyToSerialize instanceof Buffer) {
				// TODO make easy and consistent way to set headers without having to mind about casing
				responseHeaders['Content-Type'] = responseHeaders['Content-Type'] ?? 'application/octet-stream';
				res.statusMessage = responseSpec.statusMessage;
				res.writeHead(responseSpec.statusCode, responseHeaders)
				res.end(bodyToSerialize);
				return;
			}

			let encoded = Buffer.from('');
			// TODO throw not acceptable when cannot serialize

			if (typeof bodyToSerialize === 'object') {
				let serialized = '';
				if (typeof mediaType === 'undefined') {
					res.writeHead(statusCodes.HTTP_STATUS_NOT_ACCEPTABLE, errorHeaders);
					res.end();
					return;
				}
				serialized = mediaType.serialize(bodyToSerialize) as string;
				responseHeaders['Content-Type'] = mediaType.name;
				if (typeof charset !== 'undefined' && typeof responseHeaders['Content-Type'] === 'string') {
					encoded = charset.encode(serialized);
					responseHeaders['Content-Type'] += `;charset=${charset.name}`;
				}
			}

			res.statusMessage = responseSpec.statusMessage; // TODO add default status message per status code
			res.writeHead(responseSpec.statusCode, responseHeaders);
			res.end(encoded);
		} catch (errorResponseSpecRaw) {
			const responseSpec = errorResponseSpecRaw as ErrorResponse;
			res.statusMessage = responseSpec.statusMessage; // TODO add default status message per status code
			res.writeHead(responseSpec.statusCode, errorHeaders);
			res.end();
		}
	};

	serve(params: ServiceParams) {
		return new Promise<void>((resolve) => {
			this.serverInternal = new http.Server(this.requestListener);
			this.serverInternal.listen(params.port ?? 80, params.host ?? '0.0.0.0', undefined, resolve);
		});
	}

	close() {
		return new Promise<void>((resolve, reject) => {
			if (typeof this.serverInternal === 'undefined') {
				resolve();
				return;
			}
			this.serverInternal.close((err) => {
				if (err) {
					reject(err);
					return;
				}
				resolve();
			});
		});
	}
}

export const server = <Backend extends BaseBackend>(params: ServerParams<Backend>): Server<Backend> => {
	return new ServerInstance(params);
};
